import org.apache.log4j.Logger

import com.atlassian.jira.component.ComponentAccessor

import groovy.json.JsonSlurper

structureId = requestContext?.getRequestParameter("structureId")?.toLong()
structureViewId = requestContext?.getRequestParameter("structureViewId")?.toLong()

// parse parameters from the title for Better PDF Automation actions
if (!structureId) {
	try {
		def params = new JsonSlurper().parseText(title)
		structureId = params["structureId"]
		structureViewId = params["structureViewId"]
	} catch (Exception e) {
		throw new IllegalArgumentException("Failed to parse the title \"${title}\" as JSON", e)
	}
	if (structureId == null) {
		throw new IllegalArgumentException("Failed to parse \"structureId\" from the title \"${title}\"")
	}
	if (structureViewId == null) {
		throw new IllegalArgumentException("Failed to parse \"structureViewId\" from the title \"${title}\"")
	}
}

structure = new StructureTool(structureId, structureViewId)

class StructureTool {
	def log = Logger.getLogger(this.getClass())

	/**
	 * Upper limit for the items to export.
	 */
	static def MAX_ITEMS = 1000

	/**
	 * Columns to use when the structure view is unavailable.
	 */
	static def DEFAULT_COLUMNS = ["issuetype", "issuekey", "summary", "assignee", "reporter", "priority", "status", "resolution", "created", "updated", "duedate"]

	def structureId
	def structureViewId

	/* Jira components. */
	def versionManager
	def templateIssueFactory

	/* Structure components. */
	def forestService
	def i18n
	def itemResolver
	def rowManager
	def structure
	def issueKeyToIndex

	/* Structure classes. */
	def forestSpecClass
	def coreIdentitiesClass
	def coreItemTypesClass

	StructureTool(Long structureId, Long structureViewId) {
		this.structureId = structureId
		this.structureViewId = structureViewId

		versionManager = ComponentAccessor.versionManager

		def templateIssueFactoryClass = ComponentAccessor.class.classLoader.loadClass("com.atlassian.jira.mail.TemplateIssueFactory")
		templateIssueFactory = ComponentAccessor.getOSGiComponentInstanceOfType(templateIssueFactoryClass)

		def structureManager
		try {
			def structureComponentsClass = ComponentAccessor.pluginAccessor.classLoader.loadClass("com.almworks.jira.structure.api.StructureComponents")
			def structureComponents = ComponentAccessor.getOSGiComponentInstanceOfType(structureComponentsClass)
			def structureManagerClass = ComponentAccessor.pluginAccessor.classLoader.loadClass("com.almworks.jira.structure.api.structure.StructureManager")
			structureManager = ComponentAccessor.getOSGiComponentInstanceOfType(structureManagerClass)
			def structurePluginHelperClass = ComponentAccessor.pluginAccessor.classLoader.loadClass("com.almworks.jira.structure.api.StructurePluginHelper")
			def structurePluginHelper = ComponentAccessor.getOSGiComponentInstanceOfType(structurePluginHelperClass)

			forestService = structureComponents.forestService
			i18n = structurePluginHelper.i18n
			itemResolver = structureComponents.itemResolver
			rowManager = structureComponents.rowManager

			forestSpecClass = ComponentAccessor.pluginAccessor.classLoader.loadClass("com.almworks.jira.structure.api.forest.ForestSpec")
			coreIdentitiesClass = ComponentAccessor.pluginAccessor.classLoader.loadClass("com.almworks.jira.structure.api.item.CoreIdentities")
			coreItemTypesClass = ComponentAccessor.pluginAccessor.classLoader.loadClass("com.almworks.jira.structure.api.item.CoreItemTypes")
		} catch (Exception ex) {
			throw new IllegalStateException("Failed to access the Structure Java API. Installed a compatible Structure version (3.5.1 or newer)?", ex)
		}

		structure = structureManager.getStructure(structureId, null) // throws an exception if not found
	}

	def getDescription() {
		structure.description
	}

	def getName() {
		structure.name
	}

	/**
	 * Returns all issues from the structure, allowing duplicates.
	 * (When using this, make sure that the template code is prepared for duplicate issues (e.g. doesn't use issue keys as ID attributes).
	 */
	def getIssues() {
		getItems(true).collect { it.issue }
	}

	/**
	 * Returns all issues from the structure, disallowing duplicates.
	 * (In case of duplications, only the first occurrence of the duplicate issue will be returned.)
	 */
	def getUniqueIssues() {
		getIssues().unique { a, b -> a.key <=> b.key }
	}

	/**
	 * Returns all items from the structure.
	 * @param onlyIssues whether to return issue-type items only (even if so, items, not plain issues, will be returned).
	 */
	def getItems(boolean onlyIssues) {
		log.debug("Getting structure #${structureId}")
		issueKeyToIndex = [:]

		def forestSpec = forestSpecClass.structure(structureId)
		def forestSource = forestService.getForestSource(forestSpec)

		def versionedForest = forestSource.latest
		def forest = versionedForest.forest
		log.debug("Forest found: ${forest}")

		def result = []
		for (def it = 0; it < forest.size(); it++) {
			if (result.size() == MAX_ITEMS) {
				log.debug("Maximum number of items reached, exporting only ${MAX_ITEMS}")
				break
			}
			def rowId = forest.getRow(it)
			def depth = forest.getDepth(it)
			def index = forest.getPath(rowId).toList().collect {
				def siblings = forest.getPrecedingSiblings(it).toList().findAll {
					def pathItemId = rowManager.getRow(it).itemId
					!isControlItem(pathItemId)
				}
				(siblings.size() + 1).toString()
			}.join(".")
			def row = rowManager.getRow(rowId)

			def itemIdentity = row.itemId

			if (isControlItem(itemIdentity) || (onlyIssues && !this.coreIdentitiesClass.isIssue(itemIdentity))) {
				continue
			}

			def resultItem = [:]
			resultItem.putAll([
					"depth"   : depth,
					"index"   : index,
					"isItem"  : true,
					"itemType": itemIdentity.itemType
			])

			if (this.coreIdentitiesClass.isIssue(itemIdentity)) {
				def issue = templateIssueFactory.getTemplateIssue(itemResolver.resolveUnchecked(itemIdentity))
				resultItem.putAll([
						"issue"  : issue,
						"key"    : issue.key,
						"summary": issue.summary
				])
				issueKeyToIndex[issue.key] = index
			} else {
				def (name, text, color, typeName) = getItemNameAndText(itemIdentity)
				resultItem.putAll([
						"key"        : "${itemIdentity.itemType.replace("com.almworks.jira.structure:type-", "").toUpperCase()}-${index}".toString(),
						"summary"    : name,
						"description": text,
						"color"      : color,
						"issuetype"  : typeName
				])
			}
			result.add(resultItem)
		}

		log.debug("${result.size()} items found")
		log.trace("Result: ${result}")

		return result
	}

	def getIndexByIssueKey(String issueKey) {
		issueKeyToIndex[issueKey]
	}

	/**
	 * Returns the current structure view's columns.
	 * Structure allows duplicate columns within views which are returned as duplicates.
	 * To remove duplicates, change <code>.flatten().minus(null)</code> to <code>.flatten().minus(null).unique()</code>.
	 */
	def getViewColumns() {
		if (structureViewId) {
			try {
				def structureViewManagerClass = ComponentAccessor.pluginAccessor.classLoader.loadClass("com.almworks.jira.structure.api.view.StructureViewManager")
				def structureViewManager = ComponentAccessor.getOSGiComponentInstanceOfType(structureViewManagerClass)
				return structureViewManager.getView(structureViewId, null).specification.columns.findAll {
					["field", "icons", "main"].contains(it.key)
				}
				.collectNested {
					if (it.key == "field") {
						it.parameters.field
					} else if (it.key == "icons") {
						it.parameters.fields
					} else if (it.key == "main") {
						"summary"
					}
				}.flatten().minus(null)
			} catch (Exception ex) {
				log.error("Failed to access the Structure Java API. Installed a compatible Structure version?", ex)
			}
		}

		log.warn("Falling back to default columns as structure view [${structureViewId}] cannot be used")
		return DEFAULT_COLUMNS
	}

	/**
	 * Returns the triplet of (name, text, color) for an item, each eventually being <code>null</code>.
	 */
	private def getItemNameAndText(itemIdentity) {
		def itemName = null
		def itemText = null
		def itemColor = null
		def itemTypeName = null

		def item = itemResolver.resolveUnchecked(itemIdentity)

		switch (itemIdentity.itemType) {
			case coreItemTypesClass.CF_OPTION:
				itemName = item.value
				break

			case coreItemTypesClass.FOLDER:
				itemName = i18n.isKeyDefined(item.name) ? i18n.getText(item.name) : item.name.substring(item.name.lastIndexOf("/") + 1)
				itemTypeName = i18n.getText("s.w.editor.inplace.type.folder")
				break

			case coreItemTypesClass.LABEL:
				itemName = item.label
				break

			case "com.almworks.jira.structure:type-memo": // FIXME replace with coreItemTypesClass.MEMO once we only support Jira 7.6+ (Structure 5.4)
				itemName = item.name
				itemText = item.description
				itemColor = item.parameters.color
				itemTypeName = i18n.getText("s.ext.type.memo.title")
				break

			case coreItemTypesClass.USER:
				itemName = item.displayName
				break

			case "com.almworks.jira.structure:type-version-name": // FIXME replace with coreItemTypesClass.VERSION_NAME once we only support Jira 7.6+ (Structure 5.3)
				// look up the version object for its name, as item.name is lower-cased
				def versions = versionManager.getVersionsByName(item.name)
				itemName = versions ? versions.first().name : "(version not found)"
				break

			case coreItemTypesClass.COMPONENT:
			case coreItemTypesClass.ISSUETYPE:
			case coreItemTypesClass.PRIORITY:
			case coreItemTypesClass.PROJECT:
			case "com.almworks.jira.structure:type-sd-request-type": // FIXME replace with coreItemTypesClass.REQUEST_TYPE once we only support Jira 7.2+ (Structure 4.3)
			case coreItemTypesClass.RESOLUTION:
			case coreItemTypesClass.SPRINT:
			case coreItemTypesClass.STATUS:
			case "com.almworks.jira.structure:type-tempo-account": // FIXME replace with coreItemTypesClass.TEMPO_ACCOUNT once we only support Jira 7.1+ (Structure 3.6)
			case coreItemTypesClass.VERSION:
				itemName = itemResolver.resolveUnchecked(itemIdentity).name
				break

			default:
				itemName = "Unsupported item type: " + itemIdentity
		}

		[itemName, itemText, itemColor, itemTypeName]
	}

	private def isControlItem(itemIdentity) {
		coreIdentitiesClass.isGenerator(itemIdentity) || coreIdentitiesClass.isLoopMarker(itemIdentity)
	}
}
